# Copyright 2004 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

from twisted.internet import reactor
from twisted.protocols.jabber import component, jid
from twisted.internet import task
from twisted.xish.domish import Element

import xdb
import session
import jabw
import disco
import register
import utils
import config
import debug



class PyTransport(component.Service):
	def __init__(self):
		debug.log("PyTransport: Service starting up")
		self.xdb = xdb.XDB(config.jid, mangle=True)
		self.registermanager = register.RegisterManager(self)
		
		# Discovery, as well as some builtin features
		self.discovery = disco.Discovery(self)
		self.discovery.addIdentity("gateway", "msn", config.name)
		self.discovery.addIdentity("conference", "text", config.name + " Chatrooms")
		self.discovery.addFeature("jabber:iq:register", self.registermanager.incomingRegisterIq)
		self.discovery.addFeature("http://jabber.org/protocol/muc", None)
		
		self.xmlstream = None
		self.sessions = {}
		self.observers = []
		
		# Groupchat ID handling
		self.lastID = 0
		self.reservedIDs = []
		
		# Display active sessions if debug mode is on
		if(config.debugOn):
			self.loop = task.LoopingCall(self.loopCall)
			self.loop.start(60.0) # call every ten seconds
		
	
	def removeMe(self):
		debug.log("PyTransport: Service shutting down")
		dic = utils.copyDict(self.sessions)
		for session in dic:
			dic[session].removeMe()
	
	def makeID(self):
		newID = "r" + str(self.lastID)
		self.lastID += 1
		if(self.reservedIDs.count(newID) > 0):
			# Ack, it's already used.. Try again
			return self.makeID()
		else:
			return newID
	
	def reserveID(self, ID):
		self.reservedIDs.append(ID)
	
	def loopCall(self):
		if(len(self.sessions) > 0):
			debug.log("Sessions:")
			for session in self.sessions:
				debug.log("\t" + self.sessions[session].jabberID)
	
	def addObserver(self, source, pattern, callback):
		debug.log("PyTransport: Adding observer \"%s\" \"%s\" \"%s\"" % (source, pattern, str(callback)))
		self.observers.append((source, pattern, callback))
		if(self.xmlstream):
			self.xmlstream.addObserver(pattern, callback)
	
	def remObserver(self, callback):
		debug.log("PyTransport: Removing observer \"%s\"" % (str(callback)))
		for observer in self.observers:
			source, pattern, checkCB = observer
			if(callback == checkCB):
				self.observers.remove(observer)
				if(self.xmlstream):
					self.xmlstream.removeObserver(pattern, callback)
				debug.log("PyTransport: Observer removed \"%s\" \"%s\" \"%s\"" % (source, pattern, str(callback)))
				break
	
	def componentConnected(self, xmlstream):
		debug.log("PyTransport: Connected to main Jabberd server")
		self.xmlstream = xmlstream
		self.xmlstream.addObserver("/iq", self.discovery.onIq)
		self.xmlstream.addObserver("/presence", self.onPresence)
		for (source, pattern, callback) in self.observers:
			if(self.sessions.has_key(source)):
				self.xmlstream.addObserver(pattern, callback)
			else:
				self.observers.remove((source, pattern, callback))
	
	def componentDisconnected(self):
		debug.log("PyTransport: Disconnected from main Jabberd server")
		self.xmlstream = None
	
	def onPresence(self, el):
		fro = el.getAttribute("from")
		froj = jid.JID(fro.lower())
		to = el.getAttribute("to")
		if(to.find('@') < 0 and (not self.sessions.has_key(froj.userhost()))): # If the presence packet is to the transport (not a user) and there isn't already a session
			if(el.getAttribute("type") in [None, ""]): # Don't create a session unless they're sending available presence
				debug.log("PyTransport: Attempting to create a new session \"%s\"" % (froj.userhost()))
				s = session.makeSession(self, froj.userhost())
				if(s):
					self.sessions[froj.userhost()] = s
					debug.log("PyTransport: New session created \"%s\"" % (froj.userhost()))
					# Hack for first presence to get through
					s.onPresence(el)
				else:
					debug.log("PyTransport: Failed to create session \"%s\"" % (froj.userhost()))
					text = "Sorry. You do not appear to be registered with this transport. Please register and try again. If you are having trouble register please contact your Jabber administrator."
					jabw.sendMessage(self, to=froj.userhost(), fro=config.jid, body=text)
			
			elif(el.getAttribute("type") != "error"):
				debug.log("PyTransport: Sending unavailable presence to non-logged in user \"%s\"" % (froj.userhost()))
				el.swapAttributeValues("from", "to")
				el.attributes["type"] = "unavailable"
				self.send(el)
				return
	


class App:
	def __init__(self):
		self.c = component.buildServiceManager(config.jid, config.secret, "tcp:%s:%s" % (config.mainServer, config.port))
		self.transportSvc = PyTransport()
		self.transportSvc.setServiceParent(self.c)
		self.c.startService()
		reactor.addSystemEventTrigger('before', 'shutdown', self.shuttingDown)
	
	def shuttingDown(self):
		self.transportSvc.removeMe()



if(__name__ == "__main__"):
	app = App()
	reactor.run()

